<?php
/******************************************************************************
* Класс для работы с .blk файлами
*  
*/

class Blk {
	
	const MAGIC = 0x46424200;

	private static $stream = NULL;
	private static $names;
	private static $strings;

	public static function parseFile( $filename ) {

		// инициализируем буффер
		if( self::$stream == NULL ){ 

			self::$stream = new DataStream();

		}

		// открываем файл
		if( self::$stream->open( $filename ) === false ) { 

			return ErrorHandler::error( "Не удалось открыть файл '{$filename}'" );

		} 

		return self::parse();

	}

	public static function parseData( $data ) {

		// инициализируем буффер
		if( self::$stream == null ){ 

			self::$stream = new DataStream();

		}

		self::$stream->set( $data ); 

		return self::parse();

	}

	private static function parse() {   

		//  проверяем сигнатуру  "BBF\x00'
		$signature = ( self::$stream->size() >= 4 ) ? self::$stream->uint32() : 0;    

		if( $signature !== self::MAGIC ) {   
			
			return self::$stream->get();
			return ErrorHandler::error( "Файл не является сжатым BLK' " );  

		}

		// проверяем версию
		$version = self::$stream->uint16();  

		if( $version != 3 ) {

			return ErrorHandler::error( "Неизвестная версия BLK '{$version}' в файле" );

		}

		// неизвестно
		$unknown = self::$stream->uint16();
		// размер данных
		$datasize = self::$stream->uint32();

		// количество байт на количество имен параметров
		// 0х4100 - 1 байт
		// 0x8100 - 2 байта
		$countsize = self::$stream->uint16();  

		if( $countsize == 0x4100 ) { 

			$namesnum = self::$stream->byte();  

		}
		else if( $countsize == 0x8100 ) { 

			$namesnum = self::$stream->uint16();  

		}
		else  { 

			return ErrorHandler::error( "Unknown names block size " . sprintf( "0x%04X", $countsize ) 
				. " at " . sprintf( "0x%08X", self::$stream->offset - 0x04 ) );  

		}

		// имена параметров
		self::$names = array();    

		for( $i = 0; $i < $namesnum; $i++ ) {  

			$len = self::$stream->byte() ; 
			$name = self::$stream->string( $len );
			$hash = self::hash( $name ); 

			while( isset( self::$names[$hash] ) ) { 

				$hash += 256;  

			} 

			self::$names[$hash] = $name;

		}

		self::$stream->align();

		// количество байт на количество строковых параметров
		// 0х4000 - 1 байт
		// 0x8000 - 2 байта
		// 0x8001 - 3 байта
		$blocksize = self::$stream->uint16();
		$countsize = self::$stream->uint16(); 

		switch( $countsize ) { 

			case 0x8002 :
			case 0x8001 :     
			case 0x8000 :
				$stringsnum = self::$stream->uint16();
				break; 

			case 0x4000 :
				$stringsnum = self::$stream->byte();
				break; 

			case 0x0000 :
				$stringsnum = 0;
				break;  

			default :
				return ErrorHandler::error( "Unknown strings block size " . sprintf( "0x%04X", $countsize ) 
					. " at " . sprintf( "0x%08X", self::$stream->offset - 0x04 ) );  

		} 

		// строковые значения
		self::$strings = array();   

		for( $i = 0; $i < $stringsnum; $i++ ) {  

			$len = self::$stream->byte();  

			// двухбайтная длина
			if( $len > 0x7F ) {  

				$len = 0x100 * ( $len - 0x80 ) + self::$stream->byte();  

			}   

			self::$strings[$i] = self::$stream->string( $len ); 

		} 

		self::$stream->align();


		$data = array();
		self::readData( $data );

		return $data;

	}

	private static function readData( &$data ) {

		$data = array();

		// количество отдельных параметров
		$paramcount = self::$stream->uint16();         

		// количество блоков параметров
		$blockscount = self::$stream->uint16(); 

		// читаем блок отдельных параметров
		self::readParams( $paramcount, $data );

		// читаем блоки
		for( $i=0; $i < $blockscount; $i++ ) {   

			self::readBlock( $data ); 

		}  
		return;   
	}

	private static function readBlock( &$data ) {  

		// имя блока
		$key = self::$stream->uint32();  
		$blockname =  self::$names[$key];

		// один блок
		if( !isset( $data[$blockname] ) ) {

			$data[$blockname] = array();
			self::readData( $data[$blockname] );

		}
		// массив значений
		else {

			// первый элемент
			if( !is_array( $data[$blockname] )
			||  !isset( $data[$blockname][0] ) ) {

				$data[$blockname] = array( $data[$blockname] );

			}

			self::readData( $data[$blockname][count( $data[$blockname] )] );

		}

		return;
		   
	}

	private static function readParams( $count, &$data )  { 

		// список параметров
		$params = array();  

		for( $i = 0; $i < $count; $i++ ) {  

			// 2 байта на ключ?
			$params[$i]['key'] = self::$stream->uint16();
			// не используется?
			$unknown = self::$stream->byte();
			// байт на тип
			$params[$i]['type'] = self::$stream->byte();   

		}

		// список значений
		for( $i = 0; $i < $count; $i++ ) { 

			if( isset( self::$names[$params[$i]['key']] ) ) { 

				$name =  self::$names[$params[$i]['key']];
				$value = self::readParam( $params[$i]['type'] );

				// одно значение
				if( !isset( $data[$name] ) ){

					$data[$name] = $value;

				}
				// массив значений
				else {

					// первый элемент
					if( !is_array( $data[$name] ) ){

						$data[$name] = array( $data[$name], $value );

					}
					// новый элемент
					else {

						$data[$name][] = $value;

					}


				}

			}  

		} 

	}

	private static function readParam( $type )  {   

		$paramvalue = "";

		switch( $type ) {

			case 0x00:
				// size (массив uint16[2]) 
				$paramvalue = self::readSizeValue();      
				break;

			case 0x01:
				// string
				$paramvalue = self::readStringValue();
				break;  

			case 0x02:
				// int
				$paramvalue = self::readIntegerValue();
				break; 

			case 0x03:
				// float
				$paramvalue = self::readFloatValue();
				break; 

			case 0x04:
				// float[2]    
				$paramvalue = self::readFloatArray( 2 );                       
				break; 

			case 0x05:
				// float(3)                           
				$paramvalue = self::readFloatArray( 3 ); 
				break;

			case 0x06:
				// float(4)                            
				$paramvalue = self::readFloatArray( 4 ); 
				break;

			case 0x07:
				$paramvalue = self::readIntegerArray( 2 ); 
				break; 

			case 0x08:
				$paramvalue = self::readIntegerArray( 2 ); 
				break;  

			case 0x0B:
				// float(12)                          
				for( $n=0; $n<12; $n++ ) {
					$value = self::$stream->float();
					$paramvalue = $paramvalue . "$value ";
				}
				$paramvalue = trim( $paramvalue );
				break;  

			case 0x09:
				// bool false
				$paramvalue = "false";
				break;

			case 0x89:
				// bool  true
				$paramvalue = "true";
				break;

			case 0x0A:
				// RGBA                                    
				$value = self::$stream->uint32();
				$paramvalue = sprintf( "%08X", $value );
				break;

			case 0x0C:
				// int64 time
				// возвращаем только до 2038 года
				$value1 = self::$stream->uint32();
				$value2 = self::$stream->uint32();
				$paramvalue = date( "r", $value1 ); // + 4294967295 * $value2;
				break; 

			case 0x3F:
				// int
				$value = self::$stream->int32();
				$paramvalue = "0x3F => " . $value;
				break;

			default:                 
				// неизвестный тип       
				return ErrorHandler::fatal( "Unknown parameter type " . sprintf( "0x%02X", $type ) 
					. " at " . sprintf( "0x%08X", self::$stream->offset ) );

		} 

		return $paramvalue;

	}

	private static function readSizeValue() {

		return array(
			self::$stream->uint16(),
			self::$stream->uint16()
		);

	}

	private static function readStringValue() {

		$id = self::$stream->uint32();  
		$value = isset( self::$strings[$id] ) ? self::$strings[$id] : "";

		return htmlspecialchars( $value );

	}

	private static function readIntegerValue() {

		return self::$stream->int32();

	}

	private static function readFloatValue() {

		return self::$stream->float();

	}

	private static function readFloatArray( $size ) {

		$value = array();

		for( $i =0; $i < $size; $i++ ){

			$value[$i] = self::$stream->float();

		}

		return $value;

	}

	private static function readIntegerArray( $size ) {

		$value = array();

		for( $i =0; $i < $size; $i++ ){

			$value[$i] = self::$stream->float();

		}

		return $value;

	}

	private static function hash( $string ) { 

		$hash = 0x05;
		$length = strlen( $string );  

		for( $i=0; $i< $length; $i++ ) { 

			$hash = ( 33 * $hash + ord( $string[$i] ) ) & 0xFF;

		}   

		return $hash; 
	}

}



?>