<?php

/**
*  Класс для работы с dxp.bin файлами
*/

class DxpBin { 
	
	const MAGIC = "DxP2";

	private $stream;
	private $dxpFile;
	private $fat;

	function __construct(){

		$this->stream = new DataStream;

	}

	public function open( $filename ) {

		$this->dxpFile = $filename;

		if( !$this->stream->open( $filename ) ){

			return ErrorHandler::error( "Не удалось открыть файл '{$filename}'" );

		}

		// проверка сигнатуры "DxP2""
		$signature = $this->stream->string( 0x04 );

		if( $signature !== self::MAGIC ) {

			return ErrorHandler::error( "Неизвестный тип файла '{$filename}'" );

		} 

		// заголовок dxp файла 
		$version = $this->stream->uint32();       // версия файла dxp (?)

		if( $version !== 2 ) {

			return ErrorHandler::error( "Неизвестная версия DXP файла '{$filename}'" );

		}

		$this->readFat();

		$this->stream->close();
		
		return $this->fat;

	} 

	public function extractData( $image ){

		if( isset( $this->fat[$image] ) ){

			$header = $this->fat[$image]['ddsx'];
			$data = "";

			$dxpFile = fopen( $this->dxpFile, "r" );			
			fseek( $dxpFile, $this->fat[$image]['offset'] );

			if( $this->fat[$image]['size'] ){

				$data = fread( $dxpFile, $this->fat[$image]['size'] );

			}

			fclose( $dxpFile );

			return $header . $data;

		}

		return false;

	}   

	public function extractFile( $image, $file ){

		$data = self::extractData( $image );

		if( $data ) {

			return file_put_contents( $file, $data );

		}

		return false;

	}

	public function expand( $path ){

		$path = FileName::normalizeDirName( $path );

		foreach( $this->fat as $image => $info ){

			$this->extractFile( $image, $path . $image . ".ddsx" );

		}

	}

	private function readFat(){

		$textureNum = $this->stream->uint32();  
		$headerSize = $this->stream->uint32(); 
		
		$header = $this->stream->get( 0x10, $headerSize );
		$this->stream->set( $header );

		$names = $this->readFileNames();
		$ddsxInfo = $this->readDdsxInfo();
		$fileInfo = $this->readFileInfo();

		$this->fat = array();

		for( $i = 0; $i < $textureNum; $i++ ){

			// кроме пустых текстур
			if( $fileInfo[$i]['size'] ) {

				$this->fat[$names[$i]] = $fileInfo[$i];
				$this->fat[$names[$i]]['ddsx'] = $ddsxInfo[$i];

			}

		}

		return true;

	}

	private function readFileNames(){

		$this->stream->rewind();
		$fileNamesOffset = $this->stream->uint32();
		$fileNamesCount = $this->stream->uint32();

		$this->stream->seek( $fileNamesOffset );
		$fileNames = array();

		for( $i = 0; $i < $fileNamesCount; $i++ ){

			$nameOffset = $this->stream->uint32();
			$this->stream->forward( 0x04 );

			$this->stream->push();
			$this->stream->seek( $nameOffset );
			$name = FileName::normalizeFileName( $this->stream->cstring() );
			$this->stream->pop();

			// после '*' иногда содержится что-то в виде '?xxxxxx'
			$fileNames[$i] = explode( '*', $name )[0];

		}

		return $fileNames;

	}

	private function readDdsxInfo(){

		$this->stream->seek( 0x10 );
		$ddsxOffset = $this->stream->uint32();
		$ddsxCount = $this->stream->uint32();

		$this->stream->seek( $ddsxOffset );
		$ddsx = array();

		for( $i = 0; $i < $ddsxCount; $i++ ){

			$ddsx[$i] = $this->stream->string( 0x20 );

		}

		return $ddsx;

	}

	private function readFileInfo(){

		$this->stream->seek( 0x20 );
		$fileInfoOffset = $this->stream->uint32();
		$fileInfoCount = $this->stream->uint32();

		$this->stream->seek( $fileInfoOffset );
		$fileInfo = array();

		for( $i = 0; $i < $fileInfoCount; $i++ ){

			$unknown = $this->stream->uint32();
			$unknown = $this->stream->uint32();
			$unknown = $this->stream->uint32();
			$dataOffset = $this->stream->uint32();
			$dataSize = $this->stream->uint32();
			$unknown = $this->stream->uint32();

			$fileInfo[$i] = array( 
				'offset' => $dataOffset,
				'size' => $dataSize
			);

		}

		return $fileInfo;

	}

}
?>